/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2005, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * Includes code and algorithms from the Zapata library.
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Custom MYSQL CDR records.
 *
 * \author Adapted by Edward Eastman <ed@dm3.co.uk> from 
 *  cdr_custom by Mark Spencer <markster@digium.com>,
 *	cdr_addon_mysql by James Sharp <jsharp@psychoses.org>, Tilghman Lesher <asterisk__cdr__cdr_mysql__200308@the-tilghman.com>, Joseph Benden <joe@thrallingpenguin.com>
 *	
 *
 * \arg See also \ref AstCDR
 *
 *
 * \ingroup cdr_drivers
 */

#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>

#include <stdlib.h>
#include <unistd.h>
#include <time.h>

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 7682 $")

#include "asterisk/channel.h"
#include "asterisk/cdr.h"
#include "asterisk/module.h"
#include "asterisk/config.h"
#include "asterisk/pbx.h"
#include "asterisk/logger.h"
#include "asterisk/utils.h"
#include "asterisk/cli.h"
#include "asterisk/options.h"

#include <mysql.h>
#include <errmsg.h>

#define DATE_FORMAT "%Y-%m-%d %T"

AST_MUTEX_DEFINE_STATIC(lock);

static char *desc = "Customizable MYSQL CDR Backend";

static char *name = "cdr-mysql-custom";

static char format[1024]="";

static char *hostname = NULL, *dbname = NULL, *dbuser = NULL, *password = NULL, *dbsock = NULL;
static int hostname_alloc = 0, dbname_alloc = 0, dbuser_alloc = 0, password_alloc = 0, dbsock_alloc = 0;
static int dbport = 0;
static int connected = 0;
static time_t connect_time = 0;
static int records = 0;
static int totalrecords = 0;
static unsigned int timeout = 0;

static MYSQL mysql;

static char cdr_mysql_custom_status_help[] =
"Usage: cdr mysql_custom status\n"
"       Shows current connection status for cdr_mysql_custom\n";

static int handle_cdr_mysql_custom_status(int fd, int argc, char *argv[])
{
	if (connected) {
		char status[256], status2[100] = "";
		int ctime = time(NULL) - connect_time;
		if (dbport)
			snprintf(status, 255, "Connected to %s@%s, port %d", dbname, hostname, dbport);
		else if (dbsock)
			snprintf(status, 255, "Connected to %s on socket file %s", dbname, dbsock);
		else
			snprintf(status, 255, "Connected to %s@%s", dbname, hostname);

		if (dbuser && *dbuser)
			snprintf(status2, 99, " with username %s", dbuser);
		if (ctime > 31536000) {
			ast_cli(fd, "%s%s for %d years, %d days, %d hours, %d minutes, %d seconds.\n", status, status2, ctime / 31536000, (ctime % 31536000) / 86400, (ctime % 86400) / 3600, (ctime % 3600) / 60, ctime % 60);
		} else if (ctime > 86400) {
			ast_cli(fd, "%s%s for %d days, %d hours, %d minutes, %d seconds.\n", status, status2, ctime / 86400, (ctime % 86400) / 3600, (ctime % 3600) / 60, ctime % 60);
		} else if (ctime > 3600) {
			ast_cli(fd, "%s%s for %d hours, %d minutes, %d seconds.\n", status, status2, ctime / 3600, (ctime % 3600) / 60, ctime % 60);
		} else if (ctime > 60) {
			ast_cli(fd, "%s%s for %d minutes, %d seconds.\n", status, status2, ctime / 60, ctime % 60);
		} else {
			ast_cli(fd, "%s%s for %d seconds.\n", status, status2, ctime);
		}
		if (records == totalrecords)
			ast_cli(fd, "  Wrote %d records since last restart.\n", totalrecords);
		else
			ast_cli(fd, "  Wrote %d records since last restart and %d records since last reconnect.\n", totalrecords, records);
		return RESULT_SUCCESS;
	} else {
		ast_cli(fd, "Not currently connected to a MySQL server.\n");
		return RESULT_FAILURE;
	}
}

static struct ast_cli_entry cdr_mysql_custom_status_cli =
	{ { "cdr", "mysql_custom", "status", NULL },
	handle_cdr_mysql_custom_status, "Show connection status of cdr_mysql_custom",
	cdr_mysql_custom_status_help, NULL };

static int load_config(int reload) 
{
	struct ast_config *cfg;
	struct ast_variable *mappingvar;
	struct ast_variable *globalvar;
	char *tmp;
	int res = -1;

	strcpy(format, "");
	if((cfg = ast_config_load("cdr_mysql_custom.conf"))) {
		mappingvar = ast_variable_browse(cfg, "mappings");
		if(reload != 1) {
			ast_mutex_lock(&lock);
		}
		while(mappingvar) {
			if (!ast_strlen_zero(mappingvar->name) && !ast_strlen_zero(mappingvar->value)) {
				if (strlen(mappingvar->value) > (sizeof(format) - 2))
					ast_log(LOG_WARNING, "Format string too long, will be truncated, at line %d\n", mappingvar->lineno);
				strncpy(format, mappingvar->value, sizeof(format) - 2);
				strcat(format,"\n");
			} else
				ast_log(LOG_NOTICE, "Mapping must have both filename and format at line %d\n", mappingvar->lineno);
			if (mappingvar->next)
				ast_log(LOG_NOTICE, "Sorry, only one mapping is supported at this time, mapping '%s' will be ignored at line %d.\n", mappingvar->next->name, mappingvar->next->lineno); 
			mappingvar = mappingvar->next;
		}

		globalvar = ast_variable_browse(cfg, "global");
		if (!globalvar) {
			/* nothing configured */
			return 0;
		}

		tmp = ast_variable_retrieve(cfg, "global", "hostname");
		if (tmp) {
			hostname = malloc(strlen(tmp) + 1);
			if (hostname != NULL) {
				hostname_alloc = 1;
				strcpy(hostname, tmp);
			} else {
				ast_log(LOG_ERROR, "Out of memory error.\n");
				return -1;
			}
		} else {
			ast_log(LOG_WARNING, "MySQL server hostname not specified.  Assuming localhost\n");
			hostname = "localhost";
		}

		tmp = ast_variable_retrieve(cfg, "global", "dbname");
		if (tmp) {
			dbname = malloc(strlen(tmp) + 1);
			if (dbname != NULL) {
				dbname_alloc = 1;
				strcpy(dbname, tmp);
			} else {
				ast_log(LOG_ERROR, "Out of memory error.\n");
				return -1;
			}
		} else {
			ast_log(LOG_WARNING, "MySQL database not specified.  Assuming asteriskcdrdb\n");
			dbname = "asteriskcdrdb";
		}

		tmp = ast_variable_retrieve(cfg, "global", "user");
		if (tmp) {
			dbuser = malloc(strlen(tmp) + 1);
			if (dbuser != NULL) {
				dbuser_alloc = 1;
				strcpy(dbuser, tmp);
			} else {
				ast_log(LOG_ERROR, "Out of memory error.\n");
				return -1;
			}
		} else {
			ast_log(LOG_WARNING, "MySQL database user not specified.  Assuming root\n");
			dbuser = "root";
		}

		tmp = ast_variable_retrieve(cfg, "global", "sock");
		if (tmp) {
			dbsock = malloc(strlen(tmp) + 1);
			if (dbsock != NULL) {
				dbsock_alloc = 1;
				strcpy(dbsock, tmp);
			} else {
				ast_log(LOG_ERROR, "Out of memory error.\n");
				return -1;
			}
		} else {
			ast_log(LOG_WARNING, "MySQL database sock file not specified.  Using default\n");
			dbsock = NULL;
		}

		tmp = ast_variable_retrieve(cfg, "global", "password");
		if (tmp) {
			password = malloc(strlen(tmp) + 1);
			if (password != NULL) {
				password_alloc = 1;
				strcpy(password, tmp);
			} else {
				ast_log(LOG_ERROR, "Out of memory error.\n");
				return -1;
			}
		} else {
			ast_log(LOG_WARNING, "MySQL database password not specified.  Assuming blank\n");
			password = "";
		}

		tmp = ast_variable_retrieve(cfg, "global", "port");
		if (tmp) {
			if (sscanf(tmp, "%d", &dbport) < 1) {
				ast_log(LOG_WARNING, "Invalid MySQL port number.  Using default\n");
				dbport = 0;
			}
		}

		tmp = ast_variable_retrieve(cfg, "global", "timeout");
		if (tmp) {
			if (sscanf(tmp,"%d", &timeout) < 1) {
				ast_log(LOG_WARNING, "Invalid MySQL timeout number.  Using default\n");
				timeout = 0;
			}
		}

		if(reload != 1)
		{
			ast_mutex_unlock(&lock);
		}
		ast_config_destroy(cfg);
		res = 0;
	} else {
		if (reload)
			ast_log(LOG_WARNING, "Failed to reload configuration file.\n");
		else
			ast_log(LOG_WARNING, "Failed to load configuration file. Module not activated.\n");
	}
	
	return res;
}



static int mysql_log(struct ast_cdr *cdr)
{
	/* Make sure we have a big enough buf */
	char sqlcmd[2048];
	struct ast_channel dummy;

	struct tm tm;
	struct timeval tv;
	int retries = 5;

	memset(sqlcmd, 0 , sizeof(sqlcmd));
	/* Not quite the first use of a static struct ast_channel, we need it so the var funcs will work */
	memset(&dummy, 0, sizeof(dummy));
	dummy.cdr = cdr;
	pbx_substitute_variables_helper(&dummy, format, sqlcmd, sizeof(sqlcmd) - 1);

	ast_mutex_lock(&lock);

db_reconnect:
	if ((!connected) && (hostname || dbsock) && dbuser && password && dbname ) {
		/* Attempt to connect */
		mysql_init(&mysql);
		/* Add option to quickly timeout the connection */
		if (timeout && mysql_options(&mysql, MYSQL_OPT_CONNECT_TIMEOUT, (char *)&timeout)!=0) {
			ast_log(LOG_ERROR, "cdr_mysql_custom: mysql_options returned (%d) %s\n", mysql_errno(&mysql), mysql_error(&mysql));
		}
		if (mysql_real_connect(&mysql, hostname, dbuser, password, dbname, dbport, dbsock, CLIENT_MULTI_STATEMENTS)) {
			connected = 1;
			connect_time = time(NULL);
			records = 0;
		} else {
			ast_log(LOG_ERROR, "cdr_mysql_custom: cannot connect to database server %s.\n", hostname);
			connected = 0;
		}
	} else {
		/* Long connection - ping the server */
		int error;
		if ((error = mysql_ping(&mysql))) {
			connected = 0;
			records = 0;
			switch (error) {
				case CR_SERVER_GONE_ERROR:
				case CR_SERVER_LOST:
					ast_log(LOG_ERROR, "cdr_mysql_custom: Server has gone away. Attempting to reconnect.\n");
					break;
				default:
					ast_log(LOG_ERROR, "cdr_mysql_custom: Unknown connection error: (%d) %s\n", mysql_errno(&mysql), mysql_error(&mysql));
			}
			retries--;
			if (retries)
				goto db_reconnect;
			else
				ast_log(LOG_ERROR, "cdr_mysql_custom: Retried to connect fives times, giving up.\n");
		}
	}

	ast_log(LOG_DEBUG, "cdr_mysql_custom: inserting a CDR record.\n");

	if (connected) {
		if (mysql_real_query(&mysql, sqlcmd, strlen(sqlcmd))) {
			ast_log(LOG_ERROR, "cdr_mysql_custom: Failed to insert into database: (%d) %s", mysql_errno(&mysql), mysql_error(&mysql));
			mysql_close(&mysql);
			connected = 0;
		} else {
			records++;
			totalrecords++;
		}
	}
	ast_mutex_unlock(&lock);
	return 0;
}

char *description(void)
{
	return desc;
}

int unload_module(void)
{
	ast_cli_unregister(&cdr_mysql_custom_status_cli);
	if (connected) {
		mysql_close(&mysql);
		connected = 0;
		records = 0;
	}
	if (hostname && hostname_alloc) {
		free(hostname);
		hostname = NULL;
		hostname_alloc = 0;
	}
	if (dbname && dbname_alloc) {
		free(dbname);
		dbname = NULL;
		dbname_alloc = 0;
	}
	if (dbuser && dbuser_alloc) {
		free(dbuser);
		dbuser = NULL;
		dbuser_alloc = 0;
	}
	if (dbsock && dbsock_alloc) {
		free(dbsock);
		dbsock = NULL;
		dbsock_alloc = 0;
	}
	if (password && password_alloc) {
		free(password);
		password = NULL;
		password_alloc = 0;
	}
	dbport = 0;
	ast_cdr_unregister(name);
	return 0;
}

int load_module(void)
{
	int res = 0;

	if (!load_config(0)) {
		res = ast_cdr_register(name, desc, mysql_log);
		if (res) {
			ast_log(LOG_ERROR, "Unable to register custom CDR handling\n");
		} else {
			res = ast_cli_register(&cdr_mysql_custom_status_cli);
		}
	}

	mysql_init(&mysql);

	if (timeout && mysql_options(&mysql, MYSQL_OPT_CONNECT_TIMEOUT, (char *)&timeout)!=0) {
		ast_log(LOG_ERROR, "cdr_mysql: mysql_options returned (%d) %s\n", mysql_errno(&mysql), mysql_error(&mysql));
	}

	if (!mysql_real_connect(&mysql, hostname, dbuser, password, dbname, dbport, dbsock, CLIENT_MULTI_STATEMENTS)) {
		ast_log(LOG_ERROR, "Failed to connect to mysql database %s on %s.\n", dbname, hostname);
		connected = 0;
		records = 0;
	} else {
		ast_log(LOG_DEBUG, "Successfully connected to MySQL database.\n");
		connected = 1;
		records = 0;
		connect_time = time(NULL);
	}

	return res;
}

int reload(void)
{
	int res = 0;

	ast_mutex_lock(&lock);

	if (connected) {
		mysql_close(&mysql);
		connected = 0;
		records = 0;
	}
	res = load_config(1);
	mysql_init(&mysql);

	if (timeout && mysql_options(&mysql, MYSQL_OPT_CONNECT_TIMEOUT, (char *)&timeout)!=0) {
		ast_log(LOG_ERROR, "cdr_mysql: mysql_options returned (%d) %s\n", mysql_errno(&mysql), mysql_error(&mysql));
	}

	if (!mysql_real_connect(&mysql, hostname, dbuser, password, dbname, dbport, dbsock, CLIENT_MULTI_STATEMENTS)) {
		ast_log(LOG_ERROR, "Failed to connect to mysql database %s on %s.\n", dbname, hostname);
		connected = 0;
		records = 0;
	} else {
		ast_log(LOG_DEBUG, "Successfully connected to MySQL database.\n");
		connected = 1;
		records = 0;
		connect_time = time(NULL);
	}

	ast_mutex_unlock(&lock);

	return res;
}

int usecount(void)
{
	return 0;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
